//BESSELK Modified Bessel functions of the second kind.


#include<stdlib.h>
#include<stdio.h>
#include<math.h>
#ifdef __TURBOC__
  #include<conio.h>
#endif
#ifndef __TURBOC__
//  #define M_PI 3.14159265359
  #include<iostream.h>
#endif
#ifdef __cplusplus

#include "ansicpp.h"

  int max (int value1, int value2);

  int max(int value1, int value2)
  {
     return ( (value1 > value2) ? value1 : value2);
  }

#endif

   double gamma(double x)   // Gamma funkce
   {
      double  stp = 2.50662827465,
         half= 0.5,
         one = 1.0,
         fpf = 5.5;
      double  cof[6] = {76.18009173,-86.50532033,24.01409822,
         -1.231739516,0.120858003e-2,-0.536382e-5};
      int i,j;
      double xx, tmp, ser, gam;
   
      i = 0;
      if (x < 1) {
         x += 1; i = 1;
      }
      xx = x-one;  tmp = xx+fpf;  tmp = (xx+half)*log(tmp)-tmp;
      ser = one;
      for (j =0;j < 6;j++) {
         xx += one;  ser += cof[j]/xx;
      }
      gam = exp(tmp+log(stp*ser));
   
      if (i == 1) gam = gam/(x-1);
   
      return(gam);
   }


   double besselk(double alpha,double x)
   {
   
      if ((x < 0) || (alpha < 0)) {
         printf("Input arguments must be real and nonnegative.");
         exit(1);
      }
   
   //
   //  Break alpha into integer and fractional parts,
   //  and initialize result array.
   //
   
      int nfirst = (int)alpha;
      int nb = nfirst + 1;
      if (nb > 1001) {
         printf("Alpha must be <= 1000.");
         exit(1);
      }
      double enu = alpha - nfirst;
   
      double *bk;
      bk = (double *) malloc((nb+1)*sizeof(double));
   
      double eps = 2.220446049250313e-16;
      double realmin =2.225073858507203e-308;
   
   //  ln2mgamma = log(2) - Euler's constant
   //  p, q - approximation for log(gamma(1+alpha))/alpha
   //                                         + Euler's constant
   //  r, s - approximation for (1-alpha*pi/sin(alpha*pi))/(2.d0*alpha)
   //  t    - approximation for sinh(y)/y
   
      double ln2mgamma = 0.11593151565841244881;
      double p[8] = {0.805629875690432845e00, 0.204045500205365151e02,
         0.157705605106676174e03, 0.536671116469207504e03,
         0.900382759291288778e03, 0.730923886650660393e03,
         0.229299301509425145e03, 0.822467033424113231e00};
      double q[8] = {0.294601986247850434e02, 0.277577868510221208e03,
         0.120670325591027438e04, 0.276291444159791519e04,
         0.344374050506564618e04, 0.221063190113378647e04,
         0.572267338359892221e03};
      double r[6] = {-0.48672575865218401848e+0, 0.13079485869097804016e+2,
         -0.10196490580880537526e+3, 0.34765409106507813131e+3,
         0.34958981245219347820e-3};
      double s[6] = {-0.25579105509976461286e+2, 0.21257260432226544008e+3,
         -0.61069018684944109624e+3, 0.42269668805777760407e+3};
      double t[6] = { 0.16125990452916363814e-9, 0.25051878502858255354e-7,
         0.27557319615147964774e-5, 0.19841269840928373686e-3,
         0.83333333333334751799e-2, 0.16666666666666666446e+0};
   
      int ncalc = 0;
      int k = 0;
      if (enu < sqrt(realmin)) {
         enu = 0;
      }
      if (enu > 0.5) {
         k = 1;
         enu = enu - 1;
      }
      double twonu = enu+enu;
      int iend = nb+k-1;
      double c = enu*enu;
      double d3 = -c;
   
      int i,j,m,itemp,mplus1;
      double d1,d2,t1,t2,f0,f1,f2,p0,q0,bk1,bk2,dm,d;
      double ratio, x2by4, wminf, twox, blpha, pom1, pom2;
   
      if (x == 0) {
         for (i=1;i<=nb;i++) {
            bk[i] = HUGE_VAL;//*ones(1,nb);
         }
         ncalc = nb;
      }
      else if (x <= 1) {
      //
      //        Calculation of p0 = gamma(1+alpha) * (2/x)^alpha
      //                       q0 = gamma(1-alpha) * (x/2)^alpha
      //
         d1 = 0;
         d2 = p[0];
         t1 = 1;
         t2 = q[0];
         for (i = 2;i <= 7;i += 2) {
            d1 = c*d1+p[i-1];
            d2 = c*d2+p[i];
            t1 = c*t1+q[i-1];
            t2 = c*t2+q[i];
         }
         d1 = enu*d1;
         t1 = enu*t1;
         f1 = log(x);
         f0 = ln2mgamma+enu*(p[7]-enu*(d1+d2)/(t1+t2))-f1;
         q0 = exp(-enu*(ln2mgamma-enu*(p[7]+enu*(d1-d2)/(t1-t2))-f1));
         f1 = enu*f0;
         p0 = exp(f1);
         d1 = r[4];
         t1 = 1;
         for (i = 1;i<=4;i++) {
            d1 = c*d1+r[i-1];
            t1 = c*t1+s[i-1];
         }
         if (fabs(f1) <= 0.5) {
            f1 = f1*f1;
            d2 = 0;
            for (i = 1;i<=6;i++) {
               d2 = f1*d2+t[i-1];
            }
            d2 = f0+f0*f1*d2;
         }
         else {
            d2 = sinh(f1)/enu;
         }
         f0 = d2-enu*d1/(t1*p0);
         if (x <= 1.e-10) {
         //
         //        Calculation of K(alpha,x) and x*K(alpha+1,x)/K(alpha,x)
         //
            bk[1] = f0+x*f0;
            bk[1] = bk[1]-x*bk[1];
            ratio = p0/f0;
            if (k != 0) {
            //
            //            Calculation of K(alpha,x) and x*K(alpha+1,x)/K(alpha,x),
            //            alpha >= 1/2
            //
               bk[1] = ratio*bk[1]/x;
               twonu = twonu+2;
               ratio = twonu;
            }
            if (nb == 1) {
               ncalc = 1;
            }
         //
         //           Calculate  K(alpha+l,x)/K(alpha+l-1,x), l = 1,:nb-1
         //
            for (i = 2;i<=nb;i++) {
               bk[i] = ratio/x;
               twonu = twonu+2;
               ratio = twonu;
            }
            ncalc = 1;
         }
         else {
         //
         //        1.0e-10 < x <= 1.0
         //
            c = 1;
            x2by4 = x*x/4;
            p0 = 0.5*p0;
            q0 = 0.5*q0;
            d1 = -1;
            d2 = 0;
            bk1 = 0;
            bk2 = 0;
            f1 = f0;
            f2 = p0;
            t1 = HUGE_VAL;
            t2 = HUGE_VAL;
            while ((fabs(t1/(((pom1=f1+bk1)>=1)?pom1:1))>eps)
                  || (fabs(t2/(((pom2=f2+bk2)>=1)?pom2:1))>eps)) {
               d1 = d1+2;
               d2 = d2+1;
               d3 = d1+d3;
               c = x2by4*c/d2;
               f0 = (d2*f0+p0+q0)/d3;
               p0 = p0/(d2-enu);
               q0 = q0/(d2+enu);
               t1 = c*f0;
               t2 = c*(p0-d2*f0);
               bk1 = bk1+t1;
               bk2 = bk2+t2;
            }
            bk1 = f1+bk1;
            bk2 = 2*(f2+bk2)/x;
            wminf = 41.8341*x+7.1075;
         }
      }
      else if (eps*x > 1) {
      //
      //        1/eps < x
      //
         bk1 = sqrt(2/(M_PI*x));
         for (i = 1;i<=nb;i++) {
            bk[i] = bk1;
         }
         ncalc = nb;
      }
      else {
      //
      //        1.0 < x <= 1/eps
      //
         twox = x+x;
         blpha = 0;
         ratio = 0;
         if (x <= 4) {
         //
         //        Calculation of K(alpha+1,x)/K(alpha,x),  1.0 <= x <= 4.0
         //
            d2 = (int)(52.0583/x+5.7607);
            m = (int)d2;
            d1 = d2+d2;
            d2 = d2-0.5;
            d2 = d2*d2;
            for (i = 2;i<=m;i++) {
               d1 = d1-2;
               d2 = d2-d1;
               ratio = (d3+d2)/(twox+d1-ratio);
            }
         //
         //           Calculation of I(alpha,x) and I(alpha+1,x) by backward
         //           recurrence and K(alpha,x) from the Wronskian
         //
            d2 = (int)(2.7782*x+14.4303);
            m = (int)d2;
            c = fabs(enu);
            d3 = c+c;
            d1 = d3-1;
            f1 = realmin;
            f0 = (2*(c+d2)/x+0.5*x/(c+d2+1))*realmin;
            for (i = 3;i<=m;i++) {
               d2 = d2-1;
               f2 = (d3+d2+d2)*f0;
               blpha = (1+d1/d2)*(f2+blpha);
               f2 = f2/x+f1;
               f1 = f0;
               f0 = f2;
            }
            f1 = (d3+2)*f0/x+f1;
            d1 = 0;
            t1 = 1;
            for (i = 1;i<=7;i++) {
               d1 = c*d1+p[i-1];
               t1 = c*t1+q[i-1];
            }
            p0 = exp(c*(ln2mgamma+c*(p[7]-c*d1/t1)-log(x)))/x;
            f2 = (c+0.5-ratio)*f1/x;
            bk1 = p0+(d3*f0-f2+f0+blpha)/(f2+f1+f0)*p0;
            bk1 = bk1*exp(-x);
            wminf = 6.4306*x+42.511;
         }
         else {
         //
         //        Calculation of K(alpha,x) and K(alpha+1,x)/K(alpha,x)
         //        by backward recurrence, for x > 4.0
         //
            dm = (int)(185.3004/x+9.3715);
            m = (int)dm;
            d2 = dm-0.5;
            d2 = d2*d2;
            d1 = dm+dm;
            for (i = 2;i<=m;i++) {
               dm = dm-1;
               d1 = d1-2;
               d2 = d2-d1;
               ratio = (d3+d2)/(twox+d1-ratio);
               blpha = (ratio+ratio*blpha)/dm;
            }
            d = sqrt(2/M_PI);
            bk1 = 1/((d+d*blpha)*sqrt(x));
            bk1 = bk1*exp(-x);
            wminf = 1.35633*(x-fabs(x-20))+84.5096;
         }
      //
      //        Calculation of K(alpha+1,x) from K(alpha,x) and
      //        K(alpha+1,x)/K(alpha,x)
      //
         bk2 = bk1+bk1*(enu+0.5-ratio)/x;
      }
      if (!ncalc) {
      //
      //        Calculation of ncalc, K(alpha+i,x), i = 0:ncalc-1,
      //        K(alpha+i,x)/K(alpha+i-1,x), i = ncalc:nb-1
      //
         ncalc = nb;
         bk[1] = bk1;
         if (iend > 0) {
            j = 2-k;
            if (j > 0) { bk[j] = bk2; }
            if (iend > 1) {
               m = iend;
               if (wminf-enu < m) { m = (int)(wminf-enu); }
               for (i = 2;i<=m;i++) {
                  t1 = bk1;
                  bk1 = bk2;
                  twonu = twonu+2;
                  bk2 = twonu/x*bk1+t1;
                  itemp = i;
                  j = j+1;
                  if (j > 0) { bk[j] = bk2; }
               }
               m = itemp;
               if (m != iend) {
                  ratio = bk2/bk1;
                  mplus1 = m+1;
                  for (i = mplus1;i<=iend;i++) {
                     twonu = twonu+2;
                     ratio = twonu/x+1/ratio;
                     j = j+1;
                     if (j > 1) {
                        bk[j] = ratio;
                     }
                     else {
                        bk2 = ratio*bk2;
                     }
                  }
                  ncalc = max(mplus1-k,1);
                  if (ncalc == 1) { bk[1] = bk2; }
               }
            }
         }
      }
      for (i = ncalc+1;i<=nb;i++) {
         bk[i] = bk[i-1]*bk[i];
      }
      return bk[nb];
   
   }

